import asyncio
import logging

from .. import deps
from ..routes.rglob import _rglob
from ..tasks import transcode_fine

logger = logging.getLogger(__name__)


async def enqueue_all_new_fine_videos_task() -> None:
    """
    Resolve dependencies and invoke the enqueue logic to schedule fine video transcodes.

    This is the canonical entry point to enqueue fine video transcodes from places
    outside of request handlers (e.g., on startup) or from routes (via background tasks).
    """
    try:
        config = deps.config()
        if not config.PROJECT_INFO_TOML_PATH.exists():
            # Skip if project file doesn't exist.
            return
        project_info = deps.project_info(config)
        scheduler = deps.scheduler()

        # get all mp4 video files that are less than config.AUTO_TRANSCODE_VIDEO_SIZE_LIMIT_MB
        base_path = project_info.data_dir
        if base_path is None:
            # Skip if data_dir not set.
            return
        result = await asyncio.to_thread(
            _rglob, base_path, pattern="**/*.mp4", stat=True
        )

        # Sort videos by size ascending
        videos = [
            base_path / entry["path"]
            for entry in sorted(result, key=lambda entry: entry["size"])
        ]

        # Create a transcode job per video.
        # The id of the job is just the filename. We assume unique video filenames
        # across the entire dataset.
        for path in videos:
            scheduler.add_job(
                transcode_fine.transcode_video_task,
                id=path.name,
                args=[path, config.FINE_VIDEO_DIR / path.name],
                executor="transcode_pool",
                # executor="debug",
                replace_existing=True,
                misfire_grace_time=None,
            )
    except Exception:
        logger.exception("Failed to enqueue fine video transcode tasks.")



import asyncio
import logging

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from .enqueue import enqueue_all_new_fine_videos_task
from .. import deps

logger = logging.getLogger(__name__)


class ConfigFileChangeHandler(FileSystemEventHandler):
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop
        super().__init__()

    def on_modified(self, event):
        project_toml_path = deps.config().PROJECT_INFO_TOML_PATH.resolve()
        if event.src_path == str(project_toml_path):
            logger.info(
                f"Config file modified: {event.src_path}. Enqueuing new fine videos."
            )
            asyncio.run_coroutine_threadsafe(
                enqueue_all_new_fine_videos_task(), self.loop
            )


def setup_config_watcher() -> Observer:
    config_instance = deps.config()
    project_toml_path = config_instance.PROJECT_INFO_TOML_PATH.resolve()
    if not project_toml_path.exists():
        project_toml_path.parent.mkdir(parents=True, exist_ok=True)
        project_toml_path.touch()
    lp_private_dir = project_toml_path.parent

    observer = Observer()
    event_handler = ConfigFileChangeHandler(asyncio.get_event_loop())
    observer.schedule(event_handler, path=str(lp_private_dir), recursive=False)
    observer.start()
    logger.info(f"Watching for changes in {project_toml_path}")
    return observer



import logging

import pandas as pd
from fastapi import APIRouter, Depends
from pydantic import BaseModel
from starlette.concurrency import run_in_threadpool

from .project import ProjectInfo
from .. import deps
from ..tasks.extract_frames import (
    extract_frames_task,
    Session,
    MVLabelFile,
    RandomMethodOptions,
    LabelFileView,
)

logger = logging.getLogger(__name__)
router = APIRouter()
from litpose_app.config import Config


@router.post("/app/v0/rpc/getFineVideoDir")
def get_fine_video_dir(config: Config = Depends(deps.config)):
    return {"path": config.FINE_VIDEO_DIR}


class LabelFileCreationRequest(BaseModel):
    """
    Request to create label file
    if it does not exist.
    """

    labelFileTemplate: str
    """Multiview projects filenames should contain a
    star to represent view name."""


class ExtractFramesRequest(BaseModel):
    labelFileCreationRequest: LabelFileCreationRequest | None = None
    session: Session

    labelFile: MVLabelFile | None = None
    """
    Client sets None when labelFileCreationRequest is present.
    Internally it will be populated once the labelFileCreationRequest is processed.
    """

    method: str
    options: RandomMethodOptions  # add more types here with union types


@router.post("/app/v0/rpc/extractFrames")
async def extract_frames(
    request: ExtractFramesRequest,
    config: Config = Depends(deps.config),
    project_info: ProjectInfo = Depends(deps.project_info),
):

    def on_progress(x):
        logger.info(f"extractFrames progress: {x}")

    if request.labelFileCreationRequest is not None:
        assert request.labelFile is None
        mvlabelfile = await run_in_threadpool(
            init_label_file,
            request.labelFileCreationRequest,
            project_info,
        )
        request.labelFile = mvlabelfile

    await run_in_threadpool(
        extract_frames_task,
        config,
        request.session,
        project_info,
        request.labelFile,
        on_progress,
        request.method,
        request.options,
    )

    return "ok"


def init_label_file(
    labelFileCreationRequest: LabelFileCreationRequest,
    project_info: ProjectInfo,
) -> MVLabelFile:
    # Map of view to label file path
    files_to_create = []
    if "*" in labelFileCreationRequest.labelFileTemplate:
        assert project_info.views and len(project_info.views) > 0
        lfviews = []
        for view in project_info.views:
            files_to_create.append(
                project_info.data_dir
                / (
                    labelFileCreationRequest.labelFileTemplate.replace("*", view)
                    + ".csv"
                )
            )
            lfviews.append(LabelFileView(csvPath=files_to_create[-1], viewName=view))
        mvlabelfile = MVLabelFile(views=lfviews)

    else:
        files_to_create.append(
            project_info.data_dir
            / (labelFileCreationRequest.labelFileTemplate + ".csv")
        )
        mvlabelfile = MVLabelFile(
            views=[LabelFileView(csvPath=files_to_create[0], viewName="unknown")]
        )

    for p in files_to_create:
        if p.exists():
            with p.open("r") as file:
                line_count = 0
                for _ in file:
                    line_count += 1
                    if line_count >= 3:
                        raise ValueError(
                            f"Label file {p} already exists and is not empty. Stopping to prevent data loss."
                        )

    # Create the DataFrame (it's the same for all files)
    assert project_info.keypoint_names is not None
    assert len(project_info.keypoint_names) > 0
    column_levels = [["scorer"], project_info.keypoint_names, ["x", "y"]]
    column_names = ["scorer", "bodyparts", "coords"]
    column_index = pd.MultiIndex.from_product(column_levels, names=column_names)
    df = pd.DataFrame([], index=[], columns=column_index)

    # Create the label files
    for p in files_to_create:
        assert p.suffix == ".csv"
        df.to_csv(p)

    return mvlabelfile



import logging
from pathlib import Path

import tomli
import tomli_w
from fastapi import APIRouter, Depends, BackgroundTasks
from pydantic import BaseModel

from litpose_app import deps
from litpose_app.config import Config

logger = logging.getLogger(__name__)

router = APIRouter()


class ProjectInfo(BaseModel):
    """Class to hold information about the project"""

    data_dir: Path | None = None
    model_dir: Path | None = None
    views: list[str] | None = None
    keypoint_names: list[str] | None = None


class GetProjectInfoResponse(BaseModel):
    projectInfo: ProjectInfo | None  # None if project info not yet initialized


class SetProjectInfoRequest(BaseModel):
    projectInfo: ProjectInfo


@router.post("/app/v0/rpc/getProjectInfo")
def get_project_info(
    project_info: ProjectInfo = Depends(deps.project_info),
) -> GetProjectInfoResponse:
    return GetProjectInfoResponse(projectInfo=project_info)


@router.post("/app/v0/rpc/setProjectInfo")
def set_project_info(
    request: SetProjectInfoRequest,
    background_tasks: BackgroundTasks,
    config: Config = Depends(deps.config),
) -> None:
    try:
        config.PROJECT_INFO_TOML_PATH.parent.mkdir(parents=True, exist_ok=True)

        project_data_dict = request.projectInfo.model_dump(
            mode="json", exclude_none=True
        )
        try:
            with open(config.PROJECT_INFO_TOML_PATH, "rb") as f:
                existing_project_data = tomli.load(f)
        except FileNotFoundError:
            existing_project_data = {}

        # Determine if data_dir changed
        old_data_dir = existing_project_data.get("data_dir")
        new_data_dir = project_data_dict.get("data_dir")

        # Merge and persist
        existing_project_data.update(project_data_dict)
        with open(config.PROJECT_INFO_TOML_PATH, "wb") as f:
            tomli_w.dump(existing_project_data, f)

        # If data_dir changed (including being set for the first time), enqueue in background
        try:
            from litpose_app.utils.enqueue import enqueue_all_new_fine_videos_task

            if new_data_dir is not None and new_data_dir != old_data_dir:
                if background_tasks is not None:
                    background_tasks.add_task(enqueue_all_new_fine_videos_task)
                else:
                    # Fallback: schedule fire-and-forget if BackgroundTasks not provided
                    import asyncio

                    asyncio.create_task(enqueue_all_new_fine_videos_task())
        except Exception:
            # Log and continue; saving project info should not fail due to enqueue trigger
            logger.exception("Failed to schedule enqueue task after data_dir change.")

        return None

    except IOError as e:
        error_message = f"Failed to write project information to file: {str(e)}"
        print(error_message)
        raise e
    except Exception as e:
        error_message = (
            f"An unexpected error occurred while saving project info: {str(e)}"
        )
        print(error_message)
        raise e



from pathlib import Path
import logging
import shutil
import subprocess


logger = logging.getLogger("apscheduler")


def transcode_video_task(input_file_path: Path, output_file_path: Path):
    transcode_file(input_file_path, output_file_path)


# --- Configuration ---
# FFmpeg options for transcoding:
# -g 1: Intra frame for every frame (Group of Pictures size 1)
# -c:v libx264: Use libx264 encoder
# -preset medium: A balance between encoding speed and compression.
#                 Options: ultrafast, superfast, veryfast, faster, fast, medium, slow, slower, veryslow.
# -crf 23: Constant Rate Factor. Lower values mean better quality and larger files (0-51, default 23).
# -c:a copy: Copy audio stream without re-encoding. If audio re-encoding is needed, change this.
# -y: Overwrite output files without asking.
FFMPEG_OPTIONS = [
    "-c:v",
    "libx264",
    "-g",
    "1",
    "-preset",
    "medium",
    "-crf",
    "23",
    "-c:a",
    "copy",
]


def check_dependencies():
    """Checks if ffmpeg and ffprobe are installed and in PATH."""
    if shutil.which("ffmpeg") is None:
        logger.error("ffmpeg is not installed or not found in PATH.")
        return False
    if shutil.which("ffprobe") is None:
        logger.error("ffprobe is not installed or not found in PATH.")
        return False
    return True


def transcode_file(
    input_file_path: Path,
    output_file_path: Path,
) -> tuple[bool, str, Path | None]:
    """
    Transcodes a single video file to have an intra frame for every frame.
    The output file will be named by inserting ".fine" before the final ".mp4"
    and placed in the specified output_dir.
    Example: "video.sec.mp4" -> "video.sec.fine.mp4"
    Returns a tuple: (success_status: bool, message: str, output_path: Path | None)
    """
    try:

        if output_file_path.exists():
            logger.debug(
                f"Output file '{output_file_path.name}' already exists. Skipping transcoding."
            )
            return True, f"Skipped (exists): {output_file_path.name}", output_file_path

        logger.debug(f"Processing: {input_file_path.name} -> {output_file_path.name}")
        output_file_path.parent.mkdir(parents=True, exist_ok=True)
        ffmpeg_cmd = [
            "ffmpeg",
            "-i",
            str(input_file_path),
            *FFMPEG_OPTIONS,
            "-y",  # Overwrite output without asking (though we check existence above)
            str(output_file_path),
        ]

        process = subprocess.Popen(
            ffmpeg_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )

        stdout, stderr = process.communicate()

        if process.returncode == 0:
            logger.info(f"Successfully transcoded: {output_file_path.name}")
            return True, f"Success: {output_file_path.name}", output_file_path
        else:
            logger.error(f"Error transcoding '{input_file_path.name}':")
            logger.error(f"FFmpeg stdout:\n{stdout}")
            logger.error(f"FFmpeg stderr:\n{stderr}")
            # Clean up partially created file on error
            if output_file_path.exists():
                try:
                    output_file_path.unlink()
                except OSError as e:
                    logger.error(
                        f"Could not remove partially created file '{output_file_path}': {e}"
                    )
            return (
                False,
                f"Error: {input_file_path.name} - FFmpeg failed (code {process.returncode})",
                None,
            )

    except Exception as e:
        logger.error(f"Error processing '{input_file_path.name}': {e}")
        return False, f"Error: {input_file_path.name} - Exception: {e}", None



from pathlib import Path

from fastapi import APIRouter, HTTPException, Query, status, Depends
from omegaconf import OmegaConf

from litpose_app import deps
from litpose_app.routes.project import ProjectInfo

router = APIRouter()


@router.get("/app/v0/getYamlFile")
def get_yaml_file(
    file_path: Path = Query(..., alias="file_path"),
    project_info: ProjectInfo = Depends(deps.project_info),
) -> dict:
    """Reads a YAML file using OmegaConf and returns it as a plain dict.

    Args:
        file_path: Relative path to the YAML file (relative to project
            data_dir)
    Returns:
        A JSON-serializable dict with the YAML contents.

    Raises:
        404 if the file is not found.
        400 if the file cannot be parsed as YAML.
    """
    # Normalize to absolute path within the container
    path = project_info.data_dir / file_path
    if not path.is_file():
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"File not found: {file_path}",
        )

    try:
        cfg = OmegaConf.load(path)
        cfg_dict = OmegaConf.to_container(cfg, resolve=True)  # convert to plain types
        assert isinstance(cfg_dict, dict)
        return cfg_dict  # FastAPI will serialize to JSON
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Failed to parse YAML: {e}",
        )



from pathlib import Path

from ..project import ProjectInfo
from ...config import Config

def session_level_config_path(
    session_key: str, project_info: ProjectInfo, config: Config
) -> Path:
    return project_info.data_dir / config.CALIBRATIONS_DIRNAME / f"{session_key}.toml"



def find_calibration_file(
    session_key: str, project_info: ProjectInfo, config: Config
) -> None | Path:
    session_level_path = session_level_config_path(session_key, project_info, config)
    if session_level_path.is_file():
        return session_level_path

    global_calibrations_path = project_info.data_dir / config.GLOBAL_CALIBRATION_PATH
    if global_calibrations_path.is_file():
        return global_calibrations_path

    return None
from fastapi import APIRouter

from . import multiview_autolabel as _multiview_autolabel
from . import save_mvframe as _save_mvframe
from . import write_multifile as _write_multifile
from . import find_label_files as _find_label_files
from . import bundle_adjust as _bundle_adjust

# Sub-route modules within the labeler package

# Aggregate router for labeler endpoints
router = APIRouter()

# Mount sub-routers
router.include_router(_write_multifile.router)
router.include_router(_save_mvframe.router)
router.include_router(_multiview_autolabel.router)
router.include_router(_find_label_files.router)
router.include_router(_bundle_adjust.router)



from pathlib import Path

from fastapi import APIRouter, HTTPException, status
from pydantic import BaseModel

router = APIRouter()


class RGlobRequest(BaseModel):
    baseDir: Path
    pattern: str
    noDirs: bool = False
    stat: bool = False


class RGlobResponseEntry(BaseModel):
    path: Path

    # Present only if request had stat=True or noDirs=True
    type: str | None

    # Present only if request had stat=True

    size: int | None
    # Creation timestamp, ISO format.
    cTime: str | None
    # Modified timestamp, ISO format.
    mTime: str | None


class RGlobResponse(BaseModel):
    entries: list[RGlobResponseEntry]
    relativeTo: Path  # this is going to be the same base_dir that was in the request.


@router.post("/app/v0/rpc/rglob")
def rglob(request: RGlobRequest) -> RGlobResponse:
    # Prevent secrets like /etc/passwd and ~/.ssh/ from being leaked.
    if not (
        request.pattern.endswith(".csv")
        or request.pattern.endswith(".mp4")
        or request.pattern.endswith(".toml")
    ):
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Only csv, mp4, toml files are supported.",
        )

    response = RGlobResponse(entries=[], relativeTo=request.baseDir)

    results = _rglob(
        str(request.baseDir),
        pattern=request.pattern,
        no_dirs=request.noDirs,
        stat=request.stat,
    )
    for r in results:
        # Convert dict to pydantic model
        converted = RGlobResponseEntry.model_validate(r)
        response.entries.append(converted)

    return response


import datetime

from wcmatch import pathlib as w


def _rglob(base_path, pattern=None, no_dirs=False, stat=False):
    """
    Needs to be performant when searching over large model directory.
    Uses wcmatch to exclude directories with extra calls to Path.is_dir.
    wcmatch includes features that may be helpful down the line.
    """
    if pattern is None:
        pattern = "**/*"
    flags = w.GLOBSTAR
    if no_dirs:
        flags |= w.NODIR
    results = w.Path(base_path).glob(
        pattern,
        flags=flags,
    )
    result_dicts = []
    for r in results:
        stat_info = r.stat() if stat else None
        is_dir = False if no_dirs else r.is_dir() if stat else None
        if no_dirs and is_dir:
            continue
        entry_relative_path = r.relative_to(base_path)
        d = {
            "path": entry_relative_path,
            "type": "dir" if is_dir else "file" if is_dir == False else None,
            "size": stat_info.st_size if stat_info else None,
            # Note: st_birthtime is more reliable for creation time on some systems
            "cTime": (
                datetime.datetime.fromtimestamp(
                    getattr(stat_info, "st_birthtime", stat_info.st_ctime)
                ).isoformat()
                if stat_info
                else None
            ),
            "mTime": (
                datetime.datetime.fromtimestamp(stat_info.st_mtime).isoformat()
                if stat_info
                else None
            ),
        }

        result_dicts.append(d)
    return result_dicts



import aniposelib.cameras
import numpy as np
from aniposelib.cameras import CameraGroup
from fastapi import APIRouter, Depends
from pydantic import BaseModel

from litpose_app import deps
from litpose_app.routes.labeler import find_calibration_file
from litpose_app.routes.project import ProjectInfo

router = APIRouter()

import logging

logger = logging.getLogger(__name__)


class Point2D(BaseModel):
    x: float
    y: float


class Point3D(BaseModel):
    x: float
    y: float
    z: float


class KPLabel(BaseModel):
    view: str
    point: Point2D


class KPProjectedLabel(BaseModel):
    view: str
    # None if there were not enough labeled views to triangulate.
    originalPoint: Point2D | None = None
    projectedPoint: Point2D | None = None
    # Reprojection error, if the point was labeled.
    reprojection_error: float | None = None


class KeypointForRequest(BaseModel):
    keypointName: str
    labels: list[KPLabel]


class KeypointForResponse(BaseModel):
    keypointName: str
    # none if there were not enough labeled views to triangulate.
    triangulatedPt: Point3D | None = None
    projections: list[KPProjectedLabel] | None = None


class GetMVAutoLabelsRequest(BaseModel):
    sessionKey: str  # name of the session with the view stripped out, used to lookup calibration files.
    keypoints: list[KeypointForRequest]


class GetMVAutoLabelsResponse(BaseModel):
    # New keypoints obtained from triangulation + reprojection.
    # Client should patch their state with these.
    keypoints: list[KeypointForResponse]


@router.post("/app/v0/rpc/getMVAutoLabels")
def get_mv_auto_labels(
    request: GetMVAutoLabelsRequest,
    project_info: ProjectInfo = Depends(deps.project_info),
    config: deps.Config = Depends(deps.config),
) -> GetMVAutoLabelsResponse:
    # Read the toml files for this session.
    camera_group_toml_path = find_calibration_file(
        request.sessionKey, project_info, config
    )
    if camera_group_toml_path is None:
        raise FileNotFoundError(
            f"Could not find calibration file for session {request.sessionKey}"
        )

    global_cg = CameraGroup.load(camera_group_toml_path)

    results = []
    for keypoint in request.keypoints:
        res = _get_mv_auto_labels_for_keypoint(
            keypoint,
            global_cg,
        )
        results.append(res)

    return GetMVAutoLabelsResponse(keypoints=results)


def warm_up_anipose():
    """
    Invokes `aniposelib.cameras.triangulate_simple` once, because first invocation is
    slow (probably due to its use of @numba.jit).
    """
    # (aniposelib uses numba to improve numpy performance)
    from numba import config

    # On my PC this defaulted to OMP which then prevents the main process from fork().
    # (fork is necessary for any multiprocessing).
    # By switching to the simplest option, "workqueue" we also get the added benefit of more consistent
    # performance: with OMP some invocations of triangulation took ~100ms, but with workqueue, consistent ~10ms.
    config.THREADING_LAYER = "workqueue"

    try:
        cam0_mat = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0]], dtype=np.float64
        )
        aniposelib.cameras.triangulate_simple(
            np.zeros((2, 2), dtype=np.float64),
            np.array([cam0_mat, cam0_mat], dtype=np.float64),
        )
        logger.debug("Successfully warmed up anipose.")
    except Exception as e:
        logger.error("Failed to warm up anipose: {}".format(e))


def _get_mv_auto_labels_for_keypoint(
    keypoint: KeypointForRequest, global_cg: CameraGroup
):
    # Anipose triangulate for each keypoint
    labeled_views = [label.view for label in keypoint.labels]
    kp_cg = global_cg.subset_cameras_names(labeled_views)
    # skip triangulation and return appropriate value if there are not enough labeled views
    if len(keypoint.labels) < 2:
        return KeypointForResponse(
            keypointName=keypoint.keypointName,
            triangulatedPt=None,
            projections=[KPProjectedLabel(view=cam.name) for cam in global_cg.cameras],
        )

    point3d = kp_cg.triangulate(
        # makes a Cx2 array.
        np.array([[label.point.x, label.point.y] for label in keypoint.labels])
    )
    reprojections = global_cg.project(point3d)
    labels_dict = {label.view: label.point for label in keypoint.labels}

    return KeypointForResponse(
        keypointName=keypoint.keypointName,
        triangulatedPt=Point3D(x=point3d[0], y=point3d[1], z=point3d[2]),
        projections=[
            KPProjectedLabel(
                view=view,
                projectedPoint=Point2D(x=proj[0][0], y=proj[0][1]),
                originalPoint=labels_dict.get(view),
                reprojection_error=(
                    None
                    if labels_dict.get(view) is None
                    else float(
                        np.linalg.norm(
                            np.array(
                                [
                                    labels_dict.get(view).x,
                                    labels_dict.get(view).y,
                                ]
                            )
                            - proj[0]
                        )
                    )
                ),
            )
            for view, proj in zip(
                (cam.name for cam in global_cg.cameras), reprojections
            )
        ],
    )



import concurrent
from pathlib import Path

import pandas as pd
from fastapi import APIRouter, Depends

from litpose_app import deps
from litpose_app.routes.project import ProjectInfo
from litpose_app.routes.rglob import RGlobRequest, rglob
from litpose_app.utils.fix_empty_first_row import fix_empty_first_row

router = APIRouter()

import logging

logger = logging.getLogger(__name__)


@router.post("/app/v0/rpc/findLabelFiles")
def find_label_files(project_info: ProjectInfo = Depends(deps.project_info)):
    """
    A separate RPC for finding label files is needed because while the client
    has access to the RGlob RPC, RGlobbing for CSVs results in too broad of a search.
    The user may have CSVs in the data directory that are not label files.

    Instead, this endpoint first performs an rglob, but then it filters for CSVs
    that have the requisite headers to be viewed in the labeler. It uses multithreading
    to make this fast.

    Label files must have 3 header rows (scorer, bodypart, and x|y).
    We don't assume any fixed set of bodyparts.
    We just look for a unique set of bodyparts that must have both x and y.
    """

    # Get the list of files that match the glob pattern.
    # This is a list of Path objects.
    candidate_label_files_relative_paths = [
        e.path
        for e in rglob(
            RGlobRequest(
                baseDir=project_info.data_dir,
                pattern="**/*.csv",
                noDirs=True,
                stat=False,
            )
        ).entries
    ]
    # For each path, read the first 3 rows of the CSV with pandas and check
    # That the headers meet the requirements. Multithreaded.

    valid_label_files_relative_paths = []
    # Using ThreadPoolExecutor for I/O bound tasks
    with concurrent.futures.ThreadPoolExecutor() as executor:
        # Submit all tasks and store future objects
        future_to_file = {
            executor.submit(
                _check_label_file_headers, rel_path, project_info.data_dir
            ): rel_path
            for rel_path in candidate_label_files_relative_paths
        }
        for future in concurrent.futures.as_completed(future_to_file):
            original_relative_path = future_to_file[future]
            try:
                result_relative_path = future.result()
                if result_relative_path:
                    valid_label_files_relative_paths.append(result_relative_path)
            except Exception as exc:
                logger.error(
                    f"{original_relative_path} generated an exception during header check: {exc}"
                )

    return {"labelFiles": valid_label_files_relative_paths}


def _check_label_file_headers(relative_file_path: Path, base_dir: Path) -> Path | None:
    """
    Checks if a given CSV file (relative path) has the required label file headers.
    Returns the relative_file_path if valid, otherwise None.
    """
    full_file_path = base_dir / relative_file_path
    try:
        # Read only the header rows to save memory and time for large files
        df = pd.read_csv(full_file_path, header=[0, 1, 2], nrows=0)
        df = fix_empty_first_row(df)

        # Check for multi-index header with exactly 3 levels (AI-gen code, not sure if this is necessary).
        if not isinstance(df.columns, pd.MultiIndex) or df.columns.nlevels != 3:
            logger.debug(
                f"Skipping {relative_file_path}: not a multi-index header or not exactly 3 levels."
            )
            return None

        # Extract unique (scorer, bodypart) pairs from the first two levels
        # Filter out columns where the third level is not 'x' or 'y' before forming pairs
        valid_columns = [col for col in df.columns if col[2] in ("x", "y")]
        if not valid_columns:
            logger.debug(
                f"Skipping {relative_file_path}: no 'x' or 'y' coordinates found in third header level."
            )
            return None

        scorer_bodypart_pairs = set([(col[0], col[1]) for col in valid_columns])

        for scorer, bodypart in scorer_bodypart_pairs:
            has_x = False
            has_y = False
            for col in valid_columns:
                if col[0] == scorer and col[1] == bodypart:
                    if col[2] == "x":
                        has_x = True
                    elif col[2] == "y":
                        has_y = True
            if not (has_x and has_y):
                logger.debug(
                    f"Skipping {relative_file_path}: bodypart '{bodypart}' under scorer '{scorer}' does not have both 'x' and 'y' coordinates."
                )
                return None

        # If all checks pass, return the relative file path as a string
        return relative_file_path
    except pd.errors.EmptyDataError:
        logger.debug(f"Skipping {relative_file_path}: CSV is empty.")
        return None
    except Exception as e:
        logger.warning(
            f"Error processing {relative_file_path}: {type(e).__name__} - {e}"
        )
        return None



import logging
from pathlib import Path

import cv2
import numpy as np
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from tqdm import tqdm

from litpose_app.config import Config
from litpose_app.utils.video import video_capture

logger = logging.getLogger(__name__)


def frame_selection_kmeans_impl(
    config: Config,
    video_path: Path,
    n_frames: int,
    beg_frame: int = 0,
    end_frame: int | None = None,
) -> list[int]:
    """
    Reads all frames, computes motion energy, clusters into n_frames clusters
    and picks a frame out of each cluster.
    """

    # read all frames, reshape, chop off unwanted portions of beginning/end
    frames = _read_all_frames(
        config=config,
        video_file=video_path,
    )
    frame_count = frames.shape[0]
    beg_frame = 0
    # leave room for context
    end_frame = (
        frame_count - config.FRAME_EXTRACT_N_CONTEXT_FRAMES
        if end_frame is None
        else end_frame
    )
    end_frame = max(beg_frame, end_frame)
    """
    beg_frame = int(float(frame_range[0]) * frame_count)
    end_frame = int(float(frame_range[1]) * frame_count) - 2  # leave room for context
    """
    batches = np.reshape(frames, (frames.shape[0], -1))[beg_frame:end_frame]

    # take temporal diffs
    logger.info("Computing motion energy")
    me = np.concatenate([np.zeros((1, batches.shape[1])), np.diff(batches, axis=0)])
    # take absolute values and sum over all pixels to get motion energy
    me = np.sum(np.abs(me), axis=1)

    # find high me frames, defined as those with me larger than nth percentile me
    prctile = 50 if frame_count < 1e5 else 75  # take fewer frames if there are many
    idxs_high_me = np.where(me > np.percentile(me, prctile))[0]
    # just use all frames if the user wants to label a large fraction of the frames
    # (helpful for very short videos)
    if len(idxs_high_me) < n_frames:
        idxs_high_me = np.arange(me.shape[0])

    # compute pca over high me frames
    logger.info("performing pca over high motion energy frames...")
    pca_obj = PCA(n_components=np.min([batches[idxs_high_me].shape[0], 32]))
    embedding = pca_obj.fit_transform(X=batches[idxs_high_me])
    del batches  # free up memory

    # cluster low-d pca embeddings
    logger.info("performing kmeans clustering...")
    _, centers = _run_kmeans(x=embedding, n_clusters=n_frames)
    # centers is initially of shape (n_clusters, n_pcs); reformat
    centers = centers.T[None, :]

    # find high me frame that is closest to each cluster center
    # embedding is shape (n_frames, n_pcs)
    # centers is shape (1, n_pcs, n_clusters)
    dists = np.linalg.norm(embedding[:, :, None] - centers, axis=1)
    # dists is shape (n_frames, n_clusters)
    idxs_prototypes_ = np.argmin(dists, axis=0)
    # now index into high me frames to get overall indices, add offset
    idxs_prototypes = idxs_high_me[idxs_prototypes_] + beg_frame

    return idxs_prototypes


def _read_all_frames(config, video_file):
    return _read_nth_frames(config=config, video_file=video_file, n=1)


def _read_nth_frames(
    config: Config,
    video_file: Path,
    n: int = 1,
) -> np.ndarray:

    # Open the video file
    with video_capture(video_file) as cap:
        frames = []
        frame_counter = 0
        frame_total = cap.get(cv2.CAP_PROP_FRAME_COUNT)
        with tqdm(total=int(frame_total)) as pbar:
            while cap.isOpened():
                # Read the next frame
                ret, frame = cap.read()
                if ret:
                    # If the frame was successfully read, then process it
                    if frame_counter % n == 0:
                        frame_resize = cv2.resize(
                            frame,
                            (
                                config.FRAME_EXTRACT_RESIZE_DIMS,
                                config.FRAME_EXTRACT_RESIZE_DIMS,
                            ),
                        )
                        frame_gray = cv2.cvtColor(frame_resize, cv2.COLOR_BGR2RGB)
                        frames.append(frame_gray.astype(np.float16))
                    frame_counter += 1
                    progress = frame_counter / frame_total * 100.0
                    # TODO progress update
                    pbar.update(1)
                else:
                    # If we couldn't read a frame, we've probably reached the end
                    break

        return np.array(frames)


def _run_kmeans(x: np.ndarray, n_clusters: int) -> tuple:
    kmeans_obj = KMeans(n_clusters, n_init="auto")
    kmeans_obj.fit(x)
    cluster_labels = kmeans_obj.labels_
    cluster_centers = kmeans_obj.cluster_centers_
    return cluster_labels, cluster_centers



import asyncio
import logging
import os
import time
from pathlib import Path

import pandas as pd
from fastapi import APIRouter, Depends
from pydantic import BaseModel, field_validator
from starlette.concurrency import run_in_threadpool

from litpose_app import deps
from litpose_app.routes.project import ProjectInfo
from litpose_app.utils.fix_empty_first_row import fix_empty_first_row

router = APIRouter()
lock = asyncio.Lock()


logger = logging.getLogger(__name__)


class Keypoint(BaseModel):
    name: str

    # null over the wire. Converts to float("nan") due to field validator below
    x: float
    y: float

    @field_validator("x", "y", mode="before")
    def _normalize_to_nan(cls, v):
        """Convert None to NaN before validation."""
        if v is None:
            return float("nan")
        return v


class SaveFrameViewRequest(BaseModel):
    csvPath: Path  # /home/user/.../CollectedData_lTop.csv
    indexToChange: str  # labeled-data/session01_left/img001.png
    changedKeypoints: list[Keypoint]

    # If you want to remove a frame, you'd add a flag here to signal that intent.
    # For now, removal is not supported.


class SaveMvFrameRequest(BaseModel):
    views: list[SaveFrameViewRequest]


@router.post("/app/v0/rpc/save_mvframe")
async def save_mvframe(
    request: SaveMvFrameRequest,
    project_info: ProjectInfo = Depends(deps.project_info),
) -> None:
    """
    Endpoint for saving a multiview frame (in a multiview labels file).
    Global lock: Only one execution allowed at a time.
    """
    async with lock:
        # Filter out views with no changed keypoints.
        request = request.model_copy()
        request.views = list(filter(lambda v: v.changedKeypoints, request.views))

        if not request.views:
            return

        # Read files multithreaded and modify dataframes in memory.
        read_df_results = await read_df_mvframe(request)

        # Write to temp files multithreaded.
        write_tmp_results = await write_df_tmp_mvframe(
            request, read_df_results, project_info.data_dir
        )

        # Rename all files (atomic for each file).
        await commit_mvframe(request, write_tmp_results, project_info.data_dir)

        await remove_from_unlabeled_sidecar_files(project_info.data_dir, request)
    return


def _modify_df(df: pd.DataFrame, changes: SaveFrameViewRequest) -> None:
    """
    Given a dataframe with multicolumn index (scorer, bodypart, coordinate (x|y)),
    Modify the keypoints in the row at changes.index specified by changes.changedKeypoints.
    If the row doesn't exist (i.e. unlabeled frame) append to end of df.
    """
    kp_names = set(map(lambda x: x.name, changes.changedKeypoints))
    changedkps_by_name = {c.name: c for c in changes.changedKeypoints}
    columns = list(
        filter(lambda x: x[1] in kp_names and (x[2] in ["x", "y"]), df.columns.values)
    )
    logger.debug(kp_names)
    logger.debug(df.columns.values)
    new_values = []
    for c in columns:
        changedkp = changedkps_by_name[c[1]]
        if c[2] == "x":
            new_values.append(changedkp.x)
        elif c[2] == "y":
            new_values.append(changedkp.y)
        else:
            raise AssertionError('columns were filtered for c[2] in ["x", "y"]')
    logger.debug(changes.indexToChange)
    logger.debug(columns)
    logger.debug(new_values)
    df.loc[changes.indexToChange, columns] = new_values


async def read_df_mvframe(request: SaveMvFrameRequest) -> list[pd.DataFrame]:
    def read_df_file_task(vr: SaveFrameViewRequest):
        df = pd.read_csv(vr.csvPath, header=[0, 1, 2], index_col=0)
        df = fix_empty_first_row(df)
        _modify_df(df, vr)
        return df

    result = []
    for v in request.views:
        r = run_in_threadpool(read_df_file_task, v)
        result.append(r)
    return await asyncio.gather(*result)


async def write_df_tmp_mvframe(
    request: SaveMvFrameRequest,
    read_df_results: list[pd.DataFrame],
    project_data_dir: Path,
) -> list[str]:
    """
    Writes the read_df_results to temporary files, prefixed with the original file name.
    """
    timestamp = time.time_ns()
    result = []

    def write_df_to_tmp_file(v: SaveFrameViewRequest, d: pd.DataFrame):
        tmp_file = v.csvPath.with_name(f"{v.csvPath.name}.{timestamp}.tmp")
        d.to_csv(tmp_file)
        return tmp_file

    for vr, df in zip(request.views, read_df_results):
        r = run_in_threadpool(write_df_to_tmp_file, vr, df)
        result.append(r)

    return await asyncio.gather(*result)


async def commit_mvframe(
    request: SaveMvFrameRequest, tmp_file_names: list[str], project_data_dir: Path
) -> None:
    """Renames temp files to their original names (atomic per file)."""

    def commit_changes():
        for vr, tmp_file_name in zip(request.views, tmp_file_names):
            os.replace(tmp_file_name, vr.csvPath)

    return await run_in_threadpool(commit_changes)


async def remove_from_unlabeled_sidecar_files(
    data_dir: Path, request: SaveMvFrameRequest
):
    """Remove the frames from the unlabeled sidecar files.

    See also: utils.mv_label_file.py for the add version of this."""
    timestamp = time.time_ns()

    def remove_task(vr: SaveFrameViewRequest):
        unlabeled_sidecar_file = vr.csvPath.with_suffix(".unlabeled")
        if not unlabeled_sidecar_file.exists():
            return
        lines = unlabeled_sidecar_file.read_text().splitlines()
        needs_save = False
        while vr.indexToChange in lines:
            needs_save = True
            lines.remove(vr.indexToChange)

        if needs_save:
            temp_file_name = f"{unlabeled_sidecar_file.name}.{timestamp}.tmp"
            temp_file_path = unlabeled_sidecar_file.parent / temp_file_name

            temp_file_path.write_text("\n".join(lines) + "\n")
            return temp_file_path
        else:
            return None

    tasks = []
    for vr in request.views:
        tasks.append(run_in_threadpool(remove_task, vr))

    results = await asyncio.gather(*tasks)

    for vr, temp_file_path in zip(request.views, results):
        if temp_file_path is not None:
            os.replace(temp_file_path, vr.csvPath.with_suffix(".unlabeled"))



import json
import subprocess
from pathlib import Path

from fastapi import APIRouter, HTTPException, status
from pydantic import BaseModel

router = APIRouter()


class FFProbeRequest(BaseModel):
    path: Path


class FFProbeResponse(BaseModel):
    codec: str
    width: int
    height: int
    fps: int
    duration: float


@router.post("/app/v0/rpc/ffprobe")
def ffprobe(request: FFProbeRequest) -> FFProbeResponse:
    if request.path.suffix != ".mp4":
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Only mp4 files are supported.",
        )

    result = run_ffprobe(str(request.path))

    response = FFProbeResponse.model_validate(result)

    return response


def run_ffprobe(video_path):
    """
    Executes ffprobe to get video metadata and parses the JSON output.

    Args:
        video_path (str): The path to the video file.

    Returns:
        dict: A dictionary containing the parsed metadata, or None if an error occurs.
    """
    command = [
        "ffprobe",
        "-v",
        "error",  # Suppress verbose output
        "-select_streams",
        "v:0",  # Select the first video stream
        "-show_entries",
        "format=duration:stream=avg_frame_rate,codec_name,width,height,display_aspect_ratio",
        "-of",
        "json",  # Output in JSON format
        video_path,
    ]

    # Initialize extracted_info with default/None values and an error key
    extracted_info = {
        "codec": None,
        "width": None,
        "height": None,
        "fps": None,
        "duration": None,
    }

    try:
        # Execute the ffprobe command
        process = subprocess.run(command, capture_output=True, text=True, check=True)

        # Parse the JSON output
        metadata = json.loads(process.stdout)

        # --- Extracting Data ---
        if "format" in metadata and "duration" in metadata["format"]:
            try:
                extracted_info["duration"] = float(metadata["format"]["duration"])
            except ValueError:
                extracted_info["error"] = (
                    f"Failed to parse duration: {metadata['format']['duration']}"
                )
                return extracted_info

        if "streams" in metadata and len(metadata["streams"]) > 0:
            video_stream = metadata["streams"][
                0
            ]  # Assuming we want the first video stream

            # Codec
            if "codec_name" in video_stream:
                extracted_info["codec"] = str(video_stream["codec_name"])

            # Width and Height
            if "width" in video_stream:
                try:
                    extracted_info["width"] = int(video_stream["width"])
                except ValueError:
                    extracted_info["error"] = (
                        f"Failed to parse width: {video_stream['width']}"
                    )
                    return extracted_info
            if "height" in video_stream:
                try:
                    extracted_info["height"] = int(video_stream["height"])
                except ValueError:
                    extracted_info["error"] = (
                        f"Failed to parse height: {video_stream['height']}"
                    )
                    return extracted_info

            # FPS
            if "avg_frame_rate" in video_stream:
                rate_str = video_stream["avg_frame_rate"]
                if "/" in rate_str:
                    try:
                        num, den = map(int, rate_str.split("/"))
                        if den != 0:  # Avoid division by zero
                            extracted_info["fps"] = round(num / den)
                        else:
                            extracted_info["error"] = (
                                f"Framerate denominator is zero: {rate_str}"
                            )
                            return extracted_info
                    except ValueError:
                        extracted_info["error"] = (
                            f"Failed to parse fractional framerate: {rate_str}"
                        )
                        return extracted_info
                else:
                    try:
                        extracted_info["fps"] = round(float(rate_str))
                    except ValueError:
                        extracted_info["error"] = (
                            f"Failed to parse float framerate: {rate_str}"
                        )
                        return extracted_info

        return extracted_info

    except FileNotFoundError:
        extracted_info["error"] = (
            "ffprobe command not found. Please ensure FFmpeg (which includes ffprobe) is installed and accessible in your system's PATH."
        )
        return extracted_info
    except subprocess.CalledProcessError as e:
        extracted_info["error"] = (
            f"ffprobe command failed with exit code {e.returncode}. "
            f"Command: {' '.join(command)}. "
            f"Stdout: {e.stdout.strip()}. "
            f"Stderr: {e.stderr.strip()}"
        )
        return extracted_info
    except json.JSONDecodeError as e:
        extracted_info["error"] = (
            f"Failed to parse JSON output from ffprobe. Error details: {e}. "
            f"ffprobe stdout (attempted JSON): {process.stdout.strip()}"
        )
        return extracted_info
    except Exception as e:
        extracted_info["error"] = f"An unexpected error occurred: {e}"
        return extracted_info



from contextlib import contextmanager
from pathlib import Path

import cv2


@contextmanager
def video_capture(video_path: Path):
    """
    Automates checking for file existence and releasing the video capture object.
    """
    assert video_path.is_file()
    cap = cv2.VideoCapture(str(video_path))

    if not cap.isOpened():
        raise RuntimeError(f"Error opening video file {video_path}")

    try:
        yield cap
    finally:
        cap.release()



import json
import logging
import os
import subprocess
import time
from pathlib import Path
from typing import Optional
import sys  # Import sys to check the platform
import portalocker  # Import the portalocker library
import psutil  # Import the psutil library

from . import deps
from .routes.models import TrainStatus

logger = logging.getLogger(__name__)


def _is_pid_alive(pid: int) -> bool:
    """Checks if a process with the given PID is still alive.

    Uses psutil for a cross-platform check. On Unix-like systems, it explicitly
    checks for and attempts to reap zombie processes.
    """
    if pid is None:
        return False

    try:
        process = psutil.Process(pid)

        # On Unix-like systems, explicitly check for and handle zombie processes.
        # psutil.is_running() might return True for zombies in recent versions,
        # so we check status directly.
        if sys.platform != "win32":
            if process.status() == psutil.STATUS_ZOMBIE:
                logger.debug(f"Detected zombie process {pid}")
                try:
                    # Attempt to reap the zombie process.
                    # os.waitpid will clean up the zombie entry from the process table.
                    res_pid, status = os.waitpid(pid, os.WNOHANG)
                    if res_pid == pid:
                        logger.debug(f"Reaped zombie process {pid}")
                except ChildProcessError:
                    # This PID is not a child of the current process, so we can't reap it directly.
                    # It has likely been adopted by init. We still consider it not "alive" for our purpose.
                    pass
                except ProcessLookupError:
                    # PID already gone (e.g., reaped by another process or concurrent call).
                    pass
                except Exception as e:
                    logger.warning(
                        f"Error attempting to reap zombie process {pid}: {e}"
                    )
                return False  # A zombie process is not considered "alive" for active tasks.

        # If not a zombie (or on Windows where zombies don't exist), check if it's running.
        return process.is_running()

    except psutil.NoSuchProcess:
        # The PID does not exist at all, or it has terminated and been fully reaped.
        # On Unix, we can optionally attempt a final reap just in case, though it's less likely needed here.
        if sys.platform != "win32":
            try:
                res_pid, status = os.waitpid(pid, os.WNOHANG)
                if res_pid == pid:
                    logger.debug(f"Reaped terminated process {pid} after NoSuchProcess")
            except (ChildProcessError, ProcessLookupError):
                pass
            except Exception as e:
                logger.warning(f"Error during post-NoSuchProcess reap for {pid}: {e}")
        return False
    except psutil.AccessDenied:
        # Process exists but we don't have sufficient permissions to inspect it.
        # In this scenario, it's safer to assume the process is still active
        # to avoid accidentally launching another training session for it.
        logger.warning(
            f"Access denied when checking PID {pid} with psutil. "
            "Assuming process is alive to prevent double-launch."
        )
        return True
    except Exception as e:
        logger.exception(f"Unexpected error when checking PID {pid} with psutil: {e}")
        return False


def _read_status(path: Path) -> Optional[TrainStatus]:
    try:
        data = path.read_text()
        return TrainStatus.model_validate(json.loads(data))
    except Exception:
        return None


def _write_status(path: Path, status: TrainStatus) -> None:
    tmp = path.with_suffix(".tmp")
    with open(tmp, "x") as f:
        json.dump(status.model_dump(), f, indent=2)
    tmp.replace(path)


def _launch_training(model_dir: Path) -> None:
    """Launch a dummy training script that updates status and writes logs.

    In real usage, this would call the actual training script.
    """
    config_path = model_dir / "config.yaml"
    status_path = model_dir / "train_status.json"
    stdout_path = model_dir / "train_stdout.log"
    stderr_path = model_dir / "train_stderr.log"

    _write_status(status_path, TrainStatus(status="STARTING"))

    with open(stdout_path, "ab", buffering=0) as out, open(
        stderr_path, "ab", buffering=0
    ) as err:
        proc = subprocess.Popen(
            [
                "litpose",
                "train",
                str(config_path),
                "--output_dir",
                str(model_dir.parent),
            ],
            stdout=out,
            stderr=err,
            cwd=str(model_dir),
        )

    _write_status(status_path, TrainStatus(status="STARTED", pid=proc.pid))

    logger.info("Launched training pid=%s for %s", proc.pid, model_dir)


def train_scheduler_loop(poll_interval_seconds: float = 2.0) -> None:
    """Periodically checks for PENDING tasks and launches at most one training."""
    while True:
        lock_file = None
        try:
            config = deps.config()
            if not config.PROJECT_INFO_TOML_PATH.exists():
                continue
            project_info = deps.project_info(config)
            if (
                project_info
                and project_info.model_dir
                and project_info.model_dir.exists()
            ):
                base = project_info.model_dir
                lock_path = base / "scheduler.lock"

                try:
                    # Use portalocker.Lock to acquire an exclusive non-blocking lock
                    # mode='a' is important to avoid truncating existing lock files on some platforms
                    # timeout=0 means non-blocking
                    lock_file = portalocker.Lock(str(lock_path), mode="a", timeout=0)
                    lock_file.acquire()
                    logger.debug(f"Acquired lock on {lock_path}")
                except portalocker.exceptions.LockException:
                    logger.debug(
                        f"Another scheduler holds the lock on {lock_path}. Skipping this cycle."
                    )
                    continue  # Skip this cycle if lock cannot be acquired

                active_found = False
                for d in [p for p in base.iterdir() if p.is_dir()]:
                    status_path = d / "train_status.json"
                    ts = _read_status(status_path)
                    if (
                        ts
                        and ts.status
                        in ("STARTING", "STARTED", "TRAINING", "EVALUATING")
                        and ts.pid
                    ):
                        if _is_pid_alive(ts.pid):
                            active_found = True
                            break
                        else:
                            # Process died, update status to FAILED
                            _write_status(
                                status_path,
                                TrainStatus(status="FAILED", pid=ts.pid),
                            )
                            logger.info(
                                f"Marked {d.name} as FAILED due to defunct PID {ts.pid}"
                            )

                if not active_found:
                    pending_dirs = []
                    for d in sorted([p for p in base.iterdir() if p.is_dir()]):
                        status_path = d / "train_status.json"
                        ts = _read_status(status_path)
                        if ts and ts.status == "PENDING":
                            pending_dirs.append(d)
                    if pending_dirs:
                        _launch_training(pending_dirs[0])
                        logger.info(f"Launched training for {pending_dirs[0].name}")

        except Exception:
            logger.exception("Error in train scheduler loop")
        finally:
            # Finally also executes when `continue` from a try block.
            if lock_file:
                try:
                    lock_file.release()  # Release the lock
                    logger.debug(f"Released lock.")
                except Exception as e:
                    logger.error(f"Error releasing lock file {lock_path}: {e}")
            time.sleep(poll_interval_seconds)


def _train_scheduler_process_target():
    """Wrapper function to run train_scheduler_loop in a separate process."""
    # Configure logging for the child process.
    # This ensures logs from the child process are properly handled,
    # even if the main application's logging configuration changes.
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
    )
    child_logger = logging.getLogger(__name__)  # Use the same logger name

    child_logger.info("Train scheduler subprocess online.")
    try:
        train_scheduler_loop()
    except KeyboardInterrupt:
        # This can happen if the parent sends a SIGINT
        child_logger.info(
            "Train scheduler subprocess received KeyboardInterrupt, shutting down."
        )
    except Exception as e:
        child_logger.exception(
            "Train scheduler subprocess encountered an unhandled exception and is exiting."
        )
    finally:
        child_logger.info("Train scheduler subprocess exited.")


if __name__ == "__main__":
    _train_scheduler_process_target()



import os
import time
from concurrent.futures import ThreadPoolExecutor, Future
from pathlib import Path

from pydantic import BaseModel


class AddToUnlabeledFileView(BaseModel):
    # Path of the CSV file who's unlabeled sidecar file needs to be updated
    csvPath: Path
    # String repr of the paths to the frame to add to the unlabeled sidecar file
    # Relative to data dir.
    framePathsToAdd: list[str]


def add_to_unlabeled_sidecar_files(views: list[AddToUnlabeledFileView]):
    """Add frames to the unlabeled sidecar files."""
    timestamp = time.time_ns()

    def add_task(vr: AddToUnlabeledFileView):
        unlabeled_sidecar_file = vr.csvPath.with_suffix(".unlabeled")
        if not unlabeled_sidecar_file.exists():
            lines = []
            needs_save = True
        else:
            needs_save = False
            lines = unlabeled_sidecar_file.read_text().splitlines()

        for framePathToAdd in vr.framePathsToAdd:
            if framePathToAdd not in lines:
                needs_save = True
                lines.append(framePathToAdd)

        if needs_save:
            temp_file_name = f"{unlabeled_sidecar_file.name}.{timestamp}.tmp"
            temp_file_path = unlabeled_sidecar_file.parent / temp_file_name

            temp_file_path.write_text("\n".join(lines) + "\n")
            return temp_file_path
        else:
            return None

    tasks: list[Future] = []
    with ThreadPoolExecutor() as pool:
        for vr in views:
            tasks.append(pool.submit(add_task, vr))

        results = []
        for t in tasks:
            results.append(t.result())

    for vr, temp_file_path in zip(views, results):
        if temp_file_path is not None:
            os.replace(temp_file_path, vr.csvPath.with_suffix(".unlabeled"))



import reactivex
import reactivex.subject
from apscheduler import events as e
from apscheduler.schedulers.base import BaseScheduler
from fastapi import FastAPI


def setup_active_task_registry(app: FastAPI):
    # Get APScheduler instance
    scheduler: BaseScheduler = app.state.scheduler

    subject = reactivex.subject.BehaviorSubject(0)

    app.state.num_active_transcode_tasks = subject

    _num_active_jobs = 0

    def my_listener(event):
        nonlocal _num_active_jobs
        if event.code == e.EVENT_JOB_SUBMITTED:
            _num_active_jobs += 1
        else:
            _num_active_jobs -= 1

        subject.on_next(_num_active_jobs)

    scheduler.add_listener(
        my_listener,
        e.EVENT_JOB_SUBMITTED
        | e.EVENT_JOB_EXECUTED
        | e.EVENT_JOB_ERROR
        | e.EVENT_JOB_MISSED,
    )



"""
Starlette's FileResponse does not automatically return 304 if the file hasn't changed.
Starlette's StaticFiles() does, but requires a specific directory to serve files out of.

This module takes some of the logic from StaticFiles() that supports 304 if unchanged, to
implement static file serving without the directory restriction of StaticFiles().
"""

import os
from email.utils import parsedate
from pathlib import Path

from starlette import status
from starlette.exceptions import HTTPException
from starlette.responses import FileResponse
from starlette.staticfiles import NotModifiedResponse


def file_response(request, path: Path, **kwargs):
    # Follow symlinks
    path = path.resolve()

    if not path.is_file():
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    stat_result = os.stat(path)

    response = FileResponse(path, stat_result=stat_result, **kwargs)
    if _is_not_modified(response.headers, request.headers):
        return NotModifiedResponse(response.headers)
    return response


def _is_not_modified(response_headers: dict, request_headers: dict) -> bool:
    """
    Given the request and response headers, return `True` if an HTTP
    "Not Modified" response could be returned instead.

    Copied exactly from StaticFiles.is_not_modified()
    """

    try:
        if_none_match = request_headers["if-none-match"]
        etag = response_headers["etag"]
        if etag in [tag.strip(" W/") for tag in if_none_match.split(",")]:
            return True
    except KeyError:
        pass

    try:
        if_modified_since = parsedate(request_headers["if-modified-since"])
        last_modified = parsedate(response_headers["last-modified"])
        if (
            if_modified_since is not None
            and last_modified is not None
            and if_modified_since >= last_modified
        ):
            return True
    except KeyError:
        pass

    return False



import asyncio
import functools
import logging
import multiprocessing
import sys
from contextlib import asynccontextmanager
from pathlib import Path
from textwrap import dedent

import anyio
import uvicorn
from fastapi import FastAPI, HTTPException, APIRouter, Request
from fastapi.responses import FileResponse
from starlette import status
from starlette.responses import Response
from starlette.staticfiles import StaticFiles

from . import deps
from .routes.labeler.multiview_autolabel import warm_up_anipose
from .tasks.management import setup_active_task_registry
from .train_scheduler import _train_scheduler_process_target
from .utils.config_watcher import setup_config_watcher
from .utils.enqueue import enqueue_all_new_fine_videos_task
from .utils.file_response import file_response

## Setup logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


## Configure additional things to happen on server startup and shutdown.
@asynccontextmanager
async def lifespan(app: FastAPI):
    # Start apscheduler, which is responsible for executing background tasks
    logger.info("Application startup: Initializing scheduler...")
    # Quiet down apscheduler logging. Default INFO is too verbose
    logging.getLogger("apscheduler").setLevel(logging.WARNING)
    scheduler = deps.scheduler()
    app.state.scheduler = scheduler
    scheduler.start()
    setup_active_task_registry(app)

    # Kick off background task to enqueue any fine transcodes on startup
    asyncio.create_task(enqueue_all_new_fine_videos_task())

    # Setup watchdog for config file changes
    app.state.config_file_observer = setup_config_watcher()

    # Warm up anipose in the background (first run is ~1-2s slow).
    asyncio.create_task(anyio.to_thread.run_sync(warm_up_anipose))

    # Start model train scheduler loop in a separate process
    try:
        logger.info("Starting train scheduler in a separate process...")
        _train_scheduler_process = multiprocessing.Process(
            target=_train_scheduler_process_target, daemon=True
        )
        _train_scheduler_process.start()
        logger.info(f"Started train scheduler process [{_train_scheduler_process.pid}]")
    except Exception:
        logger.exception("Failed to start train scheduler process")

    yield  # Application is now ready to receive requests

    logger.info("Application shutdown: Shutting down scheduler...")
    if scheduler and scheduler.running:
        scheduler.shutdown()
        logger.info("Scheduler shut down.")
    else:
        logger.warning("Scheduler not found or not running during shutdown.")

    if hasattr(app.state, "config_file_observer") and app.state.config_file_observer:
        logger.info("Application shutdown: Shutting down config file observer...")
        app.state.config_file_observer.stop()
        app.state.config_file_observer.join()
        logger.info("Config file observer shut down.")


app = FastAPI(lifespan=lifespan)

router = APIRouter()
from .routes import (
    ffprobe,
    rglob,
    project,
    transcode,
    labeler,
    extract_frames,
    configs,
    models,
)

router.include_router(ffprobe.router)
router.include_router(rglob.router)
router.include_router(labeler.router)
router.include_router(project.router)
router.include_router(transcode.router)
router.include_router(extract_frames.router)
router.include_router(configs.router)
router.include_router(models.router)
app.include_router(router)


@app.exception_handler(Exception)
async def debug_exception_handler(request: Request, exc: Exception):
    """Puts error stack trace in response when any server exception occurs.

    By default, FastAPI returns 500 "internal server error" on any Exception
    that is not a subclass of HttpException. This is usually recommended in production apps.

    In our app, it's more convenient to expose exception details to the user. The
    security risk is minimal."""
    import traceback

    return Response(
        status_code=500,
        content="".join(
            traceback.format_exception(type(exc), value=exc, tb=exc.__traceback__)
        ),
        headers={"Content-Type": "text/plain"},
    )


"""
All our methods are RPC style (http url corresponds to method name).
They should be POST requests, /rpc/<method_name>.
Request body is some object (pydantic model).
Response body is some object pydantic model.

The client expects all RPC methods to succeed. If any RPC doesn't
return the expected response object, it will be shown as an
error in a dialog to the user. So if the client is supposed to
handle the error in any way, for example, special form validation UX
like underlining the invalid field,
then the information about the error should be included in a valid
response object rather than raised as a python error.
"""

"""
File server to serve csv and video files.
FileResponse supports range requests for video buffering.
For security - only supports reading out of data_dir and model_dir
If we need to read out of other directories, they should be added to Project Info.
"""


@app.get("/app/v0/files/{file_path:path}")
async def read_file(request: Request, file_path: Path):
    # Prevent secrets like /etc/passwd and ~/.ssh/ from being leaked.
    if file_path.suffix not in (".csv", ".mp4", ".png", ".jpg", ".unlabeled", ".log"):
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="File type not supported: " + file_path.suffix,
        )
    file_path = Path("/") / file_path

    # Prevent browser caching of data files (video and image caching are fine, (for now)).
    # no-cache: browser must check if its cached value is still valid (but can still use its cached value if so).
    headers = (
        {"Cache-Control": "no-cache"}
        if file_path.suffix not in (".mp4", ".png", ".jpg")
        else None
    )
    partial = functools.partial(file_response, request, file_path, headers=headers)
    response = await anyio.to_thread.run_sync(partial)
    return response


###########################################################################
# Serving angular
#
# In dev mode, `ng serve` serves ng, and proxies to us for backend requests.
# In production mode, we will serve ng.
# This is necessary to use HTTP2 for faster concurrent request performance (ng serve doesn't support it).
###########################################################################


# Serve ng assets (js, css)
STATIC_DIR = Path(__file__).parent / "ngdist" / "ng_app" / "browser"
if not STATIC_DIR.is_dir():
    message = dedent(
        """
        ⚠️  Warning: We couldn't find the necessary static assets (like HTML, CSS, JavaScript files).
        As a result, only the HTTP API is currently running.

        This usually happens if you've cloned the source code directly.
        To fix this and get the full application working, you'll need to either:

        - Build the application: Refer to development.md in the repository for steps.
        - Copy static files: Obtain these files from a PyPI source distribution of a released
        version and place them in:

            {STATIC_DIR}
        """
    )
    # print(f'{Fore.white}{Back.yellow}{message}{Style.reset}', file=sys.stderr)
    print(f"{message}", file=sys.stderr)

app.mount("/static", StaticFiles(directory=STATIC_DIR, check_dir=False), name="static")


@app.get("/favicon.ico")
async def favicon():
    return FileResponse(
        Path(__file__).parent / "ngdist" / "ng_app" / "browser" / "favicon.ico"
    )


# Catch-all route. serve index.html.
@app.get("/{full_path:path}")
async def index():
    return FileResponse(
        Path(__file__).parent / "ngdist" / "ng_app" / "browser" / "index.html"
    )


def run_app(host: str, port: int):
    uvicorn.run(app, host=host, port=port, timeout_graceful_shutdown=1)



import asyncio
import json
import logging
from datetime import datetime
from typing import Literal
from pathlib import Path

import yaml

from fastapi import APIRouter, Depends, HTTPException, status
from pydantic import BaseModel, Field

from litpose_app import deps
from litpose_app.routes.project import ProjectInfo

logger = logging.getLogger(__name__)

router = APIRouter()


StatusLiteral = Literal[
    "PENDING",
    "STARTING",
    "STARTED",
    "TRAINING",
    "EVALUATING",
    "COMPLETED",
    "FAILED",
    "CANCELED",
    "PAUSED",
]


class TrainStatus(BaseModel):
    status: StatusLiteral
    pid: int | None = None


class CreateTrainTaskRequest(BaseModel):
    modelName: str = Field(..., min_length=1)
    # YAML as string, but we store it verbatim; client may send object -> we will stringify if needed
    configYaml: str


class CreateTrainTaskResponse(BaseModel):
    ok: bool


class ModelListResponseEntry(BaseModel):
    model_name: str
    model_relative_path: str
    config: dict | None
    created_at: str  # ISO format
    status: TrainStatus | None = None


class ListModelsResponse(BaseModel):
    models: list[ModelListResponseEntry]


@router.post("/app/v0/rpc/createTrainTask")
def create_train_task(
    request: CreateTrainTaskRequest,
    project_info: ProjectInfo = Depends(deps.project_info),
) -> CreateTrainTaskResponse:
    if project_info is None or project_info.model_dir is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Project model_dir is not configured.",
        )

    model_dir = Path(project_info.model_dir / request.modelName).resolve()

    # Ensure model name maps within model_dir
    try:
        model_dir.relative_to(project_info.model_dir)
    except Exception:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Invalid model name.",
        )

    if model_dir.exists():
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail="Model already exists.",
        )

    model_dir.mkdir(parents=True, exist_ok=False)

    # Save config.yaml
    (model_dir / "config.yaml").write_text(request.configYaml)

    # Create initial train_status.json
    status_path = model_dir / "train_status.json"
    status_json = TrainStatus(status="PENDING").model_dump()
    status_path.write_text(json.dumps(status_json, indent=2))

    # Prepare stdout/stderr files for future training
    (model_dir / "train_stdout.log").touch(exist_ok=True)
    (model_dir / "train_stderr.log").touch(exist_ok=True)

    return CreateTrainTaskResponse(ok=True)


@router.post("/app/v0/rpc/listModels")
def list_models(
    project_info: ProjectInfo = Depends(deps.project_info),
) -> ListModelsResponse:
    models: list[ModelListResponseEntry] = []
    if project_info is None or project_info.model_dir is None:
        return ListModelsResponse(models=models)

    base = Path(project_info.model_dir)
    if not base.exists():
        return ListModelsResponse(models=models)

    def read_model_config(child_path: Path) -> ModelListResponseEntry:
        config_path = child_path / "config.yaml"
        status_path = child_path / "train_status.json"
        config = None
        status = None

        if config_path.is_file():
            try:
                content = config_path.read_text()
                config = yaml.safe_load(content)
            except Exception:
                logger.exception("Failed to read config.yaml for %s", child_path)

        if status_path.is_file():
            try:
                content = status_path.read_text()
                status_data = json.loads(content)
                status = TrainStatus(**status_data)
            except Exception:
                logger.exception("Failed to read train_status.json for %s", child_path)

        stat = child_path.stat()
        created_at = datetime.fromtimestamp(stat.st_ctime).isoformat()

        return ModelListResponseEntry(
            model_name=child_path.name,
            model_relative_path=str(child_path.relative_to(base)),
            config=config,
            created_at=created_at,
            status=status,
        )

    paths = sorted([p for p in base.iterdir() if p.is_dir()])
    models = [read_model_config(p) for p in paths]

    return ListModelsResponse(models=models)



"""
Dependencies that can be injected into routes.
This has the benefit of making tests easier to write, as you can override dependencies.
See FastAPI Dependency Injection docs: https://fastapi.tiangolo.com/tutorial/dependencies/
"""

from __future__ import annotations

import logging
import math
import os
from typing import TYPE_CHECKING

from apscheduler.executors.debug import DebugExecutor
from apscheduler.executors.pool import ThreadPoolExecutor
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from fastapi import Depends

from litpose_app.config import Config

logger = logging.getLogger(__name__)


def config() -> Config:
    """Dependency that provides the app config object."""
    from .main import app

    if not hasattr(app.state, "config"):
        app.state.config = Config()
    return app.state.config


def scheduler() -> AsyncIOScheduler:
    """Dependency that provides the app's APScheduler instance."""
    from .main import app

    if not hasattr(app.state, "scheduler"):
        # ffmpeg parallelizes transcoding to the optimal degree, but
        # that doesn't always saturate a machine with a lot of cores.
        # i.e. on a 24 logical core machine (12 physical * 2 hyperthreads per core)
        # 3 was the ideal number of max_workers. Let's just guesstimate that
        # ffmpeg uses 10 cores? No scientific evidence, but ceil(24/10) => 3.
        transcode_workers = math.ceil(os.cpu_count() / 10)
        executors = {
            "transcode_pool": ThreadPoolExecutor(max_workers=transcode_workers),
            "debug": DebugExecutor(),
        }
        app.state.scheduler = AsyncIOScheduler(executors=executors)
    return app.state.scheduler


if TYPE_CHECKING:
    from .routes.project import ProjectInfo


def project_info(config: Config = Depends(config)) -> ProjectInfo:
    import tomli
    from .routes.project import ProjectInfo

    from pydantic import ValidationError

    try:
        # Open the file in binary read mode, as recommended by tomli
        with open(config.PROJECT_INFO_TOML_PATH, "rb") as f:
            # Load the TOML data into a Python dictionary
            toml_data = tomli.load(f)

        # Unpack the dictionary into the Pydantic model
        return ProjectInfo.model_validate(toml_data)
    except FileNotFoundError:
        raise RuntimeError("project not yet setup, but project_info dep requested")
    except tomli.TOMLDecodeError as e:
        logger.error(f"Could not decode pyproject.toml. Invalid syntax: {e}")
        raise
    except ValidationError as e:
        logger.error(f"pyproject.toml is invalid. {e}")
        raise



import asyncio
from pathlib import Path

import aiofiles
import aiofiles.os
from fastapi import APIRouter, Depends
from pydantic import BaseModel

from litpose_app import deps
from litpose_app.routes.project import ProjectInfo


class FileToWrite(BaseModel):
    filename: str
    contents: str


class WriteMultifileRequest(BaseModel):
    views: list[FileToWrite]


router = APIRouter()


@router.post("/app/v0/rpc/writeMultifile")
async def write_multifile(
    request: WriteMultifileRequest,
    project_info: ProjectInfo = Depends(deps.project_info),
):
    # Security
    for view in request.views:
        p = Path(view.filename).resolve()  # crucial to resolve ".."
        if not p.is_relative_to(project_info.data_dir):
            raise AssertionError("Invalid filename")
        if p.suffix not in (".csv", ".unlabeled"):
            raise AssertionError("Invalid suffix")

    # Write all files to tmpfile to ensure they all successfully write.
    write_tasks = []
    for view in request.views:
        tmpfile = view.filename + f".lptmp"

        async def write_file_task(filename, contents):
            async with aiofiles.open(filename, mode="w") as f:
                await f.write(contents)

        write_tasks.append(write_file_task(tmpfile, view.contents))

    await asyncio.gather(*write_tasks)

    # Rename is atomic. Partial failure is highly unlikely since all writes above succeeded.
    # In case of partial failure, the remaining tmpfiles created above aid investigation.
    rename_tasks = []
    for view in request.views:
        tmpfile = view.filename + ".lptmp"
        rename_tasks.append(aiofiles.os.rename(tmpfile, view.filename))

    await asyncio.gather(*rename_tasks)

    return "ok"



import numpy as np
import pandas as pd


def fix_empty_first_row(df: pd.DataFrame) -> pd.DataFrame:
    """Copied from lightning-pose utils. Fixes pandas issue where reading a dataframe with NaNs in first line omits it."""
    if df.index.name is not None:
        new_row = {col: np.nan for col in df.columns}
        prepend_df = pd.DataFrame(
            new_row, index=[df.index.name], columns=df.columns, dtype="float64"
        )
        fixed_df = pd.concat([prepend_df, df])
        assert fixed_df.index.name is None
        return fixed_df

    return df



import logging
from pathlib import Path

import cv2
import numpy as np

from litpose_app.config import Config
from litpose_app.utils.video import video_capture

logger = logging.getLogger(__name__)


def export_frames_singleview_impl(
    config: Config, video_path: Path, frame_idxs: np.ndarray, dest_paths: list[Path]
):
    with video_capture(video_path) as cap:
        frames = get_frames_from_idxs(cap, frame_idxs)
    for frame, dest_path in zip(frames, dest_paths):
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        cv2.imwrite(
            filename=str(dest_path),
            img=cv2.cvtColor(frame.transpose(1, 2, 0), cv2.COLOR_RGB2BGR),
        )


def get_frames_from_idxs(cap: cv2.VideoCapture, idxs: np.ndarray) -> np.ndarray:
    """Helper function to load video segments.

    Parameters
    ----------
    cap : cv2.VideoCapture object
    idxs : array-like
        frame indices into video

    Returns
    -------
    np.ndarray
        returned frames of shape (n_frames, n_channels, ypix, xpix)

    """
    is_contiguous = np.sum(np.diff(idxs)) == (len(idxs) - 1)
    n_frames = len(idxs)
    for fr, i in enumerate(idxs):
        if fr == 0 or not is_contiguous:
            cap.set(1, i)
        ret, frame = cap.read()
        if ret:
            frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            if fr == 0:
                height, width, _ = frame_rgb.shape
                frames = np.zeros((n_frames, 3, height, width), dtype="uint8")
            frames[fr] = frame_rgb.transpose(2, 0, 1)
        else:
            logger.error(
                "Reached end of video; returning blank frames for remainder of requested indices"
            )
            break
    return frames



"""Routes should not access this directly, if they want to be able to
modify these in unit tests.
Instead, prefer to inject `config: deps.config into the route using FastAPI's dependency injection.
See https://fastapi.tiangolo.com/tutorial/dependencies/."""

import os
from pathlib import Path

from pydantic import BaseModel


# Consider `pydantic_settings.BaseSettings` for potential future needs.
class Config(BaseModel):
    PROJECT_INFO_TOML_PATH: Path = Path("~/.lightning_pose/project.toml").expanduser()

    ## Video transcoding settings

    # Directory where finely transcoded videos are stored
    FINE_VIDEO_DIR: Path = Path("~/.lightning_pose/finevideos").expanduser()

    # Name of the directory in data_dir where extract frames will output to
    LABELED_DATA_DIRNAME: str = "labeled_data"

    # Name of the directory in data_dir containing session-specific calibration files
    CALIBRATIONS_DIRNAME: str = "calibrations"
    CALIBRATION_BACKUPS_DIRNAME: str = "calibration_backups"

    GLOBAL_CALIBRATION_PATH: str = "calibration.toml"

    ###
    # Frame extraction config
    ###

    FRAME_EXTRACT_N_CONTEXT_FRAMES: int = 2
    FMT_FRAME_INDEX_DIGITS: int = 8
    N_WORKERS: int = os.cpu_count()
    FRAME_EXTRACT_RESIZE_DIMS: int = 64



import asyncio
import json
import logging
from typing import AsyncGenerator

import reactivex
import reactivex.operators
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from fastapi import APIRouter, Depends, Request
from fastapi.responses import StreamingResponse
from reactivex import Observable

from .. import deps
from ..utils.enqueue import enqueue_all_new_fine_videos_task


logger = logging.getLogger(__name__)
router = APIRouter()
from litpose_app.config import Config


@router.post("/app/v0/rpc/getFineVideoDir")
def get_fine_video_dir(config: Config = Depends(deps.config)):
    return {"path": config.FINE_VIDEO_DIR}


@router.get("/app/v0/rpc/getFineVideoStatus")
async def get_fine_video_status(request: Request) -> StreamingResponse:
    """
    Returns number of pending transcode tasks.
    """
    subject: reactivex.Subject = request.app.state.num_active_transcode_tasks
    wrapped = subject.pipe(reactivex.operators.map(lambda x: {"pending": x}))
    return await sse_events(request, wrapped)


@router.post("/app/v0/rpc/enqueueAllNewFineVideos")
async def enqueue_all_new_fine_videos():

    await enqueue_all_new_fine_videos_task()
    return "ok"


async def sse_events(request: Request, source: Observable) -> StreamingResponse:
    # An asyncio.Queue will act as a bridge between the synchronous world of
    # the RxPy observable and the async world of the FastAPI response.
    queue = asyncio.Queue()

    # The scheduler ensures that the observable runs on the asyncio event loop,
    # which is necessary for it to work correctly with FastAPI.
    scheduler = AsyncIOScheduler()

    async def event_generator() -> AsyncGenerator[str, None]:
        """
        This async generator function is the core of the SSE streaming.
        It listens for items on the queue and yields them in the SSE format.
        """
        disposable = None
        try:
            # Subscribe to the observable.
            # For each item emitted by the observable (`on_next`), we put it
            # into our asyncio queue.
            # If the observable completes (`on_completed`) or errors (`on_error`),
            # we put a special sentinel value (None) in the queue to signal the end.
            disposable = source.subscribe(
                on_next=lambda item: queue.put_nowait(item),
                on_error=lambda e: queue.put_nowait(None),
                on_completed=lambda: queue.put_nowait(None),
                scheduler=scheduler,
            )

            while True:
                # Check if the client has disconnected.
                if await request.is_disconnected():
                    logger.debug("Client disconnected.")
                    break

                # Wait for an item to appear on the queue.
                item = await queue.get()

                # If the item is our sentinel value, it means the observable
                # has finished, and we can stop streaming.
                if item is None:
                    break

                # Yield the data in the Server-Sent Event format.
                # The format is "data: {your_message}\n\n"
                yield f"data: {json.dumps(item)}\n\n"

        finally:
            # This block is crucial for cleanup. It runs when the client
            # disconnects or the stream is otherwise closed.
            # We dispose of the subscription to the observable, which stops
            # the interval and prevents memory leaks.
            if disposable:
                disposable.dispose()
                logger.debug("Observable subscription disposed.")
            logger.debug("Event generator finished.")

    return StreamingResponse(event_generator(), media_type="text/event-stream")



import logging
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Callable

import cv2
import numpy as np
from pydantic import BaseModel

from litpose_app.config import Config
from litpose_app.routes.project import ProjectInfo
from litpose_app.utils.mv_label_file import (
    AddToUnlabeledFileView,
    add_to_unlabeled_sidecar_files,
)
from litpose_app.utils.video import video_capture
from litpose_app.utils.video.export_frames import export_frames_singleview_impl
from litpose_app.utils.video.frame_selection import frame_selection_kmeans_impl

logger = logging.getLogger(__name__)


class SessionView(BaseModel):
    videoPath: Path
    viewName: str


class Session(BaseModel):
    views: list[SessionView]


class LabelFileView(BaseModel):
    csvPath: Path
    viewName: str


class MVLabelFile(BaseModel):
    views: list[LabelFileView]


class RandomMethodOptions(BaseModel):
    nFrames: int = 10


DEFAULT_RANDOM_OPTIONS = RandomMethodOptions()

# Other configuration


def extract_frames_task(
    config: Config,
    session: Session,
    project_info: ProjectInfo,
    mv_label_file,
    progress_callback: Callable[[str], None],
    method="random",
    options: RandomMethodOptions = DEFAULT_RANDOM_OPTIONS,
):
    """
    session: dict (serialized Session model)
    method: random (kmeans) | active (NYI)
    """

    frame_idxs: list[int] = []
    with ProcessPoolExecutor(
        max_workers=min(config.N_WORKERS, len(session.views))
    ) as process_pool:
        if method == "random":
            frame_idxs = _frame_selection_kmeans(config, session, options, process_pool)
            progress_callback(f"Frame selection complete.")
        else:
            raise ValueError("method not supported: " + method)

        result = _export_frames(config, session, project_info, frame_idxs, process_pool)
        progress_callback(f"Frame extraction complete.")
        logger.debug(result)
        _update_unlabeled_files(project_info.data_dir, result, mv_label_file)
        progress_callback(f"Update unlabeled files complete.")


def _frame_selection_kmeans(config, session, options, process_pool) -> list[int]:
    """
    Select `options.n_frames` frames using just the first video in the session.

    Offload it to a separate process because this is CPU-intensive.
    """

    future = process_pool.submit(
        frame_selection_kmeans_impl,
        config,
        session.views[0].videoPath,
        options.nFrames,
    )
    return future.result()


def _export_frames(
    config: Config,
    session: Session,
    project_info: ProjectInfo,
    frame_idxs,
    process_pool,
) -> dict[str, list[Path]]:
    """
    Extracts frames (frame_idxs) from each view.

    Work is executed by process pool: one task per camera view.
    Tasks write to temp files.
    Upon all tasks finishing, the temp files are moved to final destination (atomic).

    Returns a dict of view_name -> list of paths to extracted center frames (relative to data dir).
    """

    def dest_path(video_path: Path, frame_idx: int) -> Path:
        return (
            project_info.data_dir
            / config.LABELED_DATA_DIRNAME
            / video_path.stem
            / f"img{frame_idx:0{config.FMT_FRAME_INDEX_DIGITS}d}.jpg"
        )

    # Compute destination paths for center frames as the return value of the function.
    retval = {
        sv.viewName: [dest_path(sv.videoPath, frame_idx) for frame_idx in frame_idxs]
        for sv in session.views
    }

    def get_frame_count(video_path: Path) -> int:
        with video_capture(video_path) as cap:
            return int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

    frame_count = get_frame_count(session.views[0].videoPath)

    # expand frame_idxs to include context frames
    context_frames = config.FRAME_EXTRACT_N_CONTEXT_FRAMES
    if context_frames > 0:
        context_vec = np.arange(-context_frames, context_frames + 1)
        _result = (frame_idxs[None, :] + context_vec[:, None]).flatten()
        _result.sort()
        _result = _result[_result >= 0]
        _result = _result[_result < int(frame_count)]
        _result = np.unique(_result)
        frame_idxs_with_context = _result

    futures = {}
    for sv in session.views:
        # Compute destination paths for every frame including context frames.
        dest_paths = [dest_path(sv.videoPath, idx) for idx in frame_idxs_with_context]
        future = process_pool.submit(
            export_frames_singleview_impl,
            config,
            sv.videoPath,
            frame_idxs_with_context,
            dest_paths,
        )
        futures[sv.viewName] = future

    # Wait for all completion
    for view_name, future in futures.items():
        future.result()

    return retval


def _update_unlabeled_files(
    data_dir: Path, result: dict[str, list[Path]], mv_label_file: MVLabelFile
):
    """
    Appends the new frames in `result` to the `mv_label_file` as atomically as possible.
    """
    lfv_dict = {lfv.viewName: lfv for lfv in mv_label_file.views}
    x = [
        AddToUnlabeledFileView(
            csvPath=lfv_dict[view_name].csvPath,
            framePathsToAdd=[str(p.relative_to(data_dir)) for p in result[view_name]],
        )
        for view_name in result
    ]
    add_to_unlabeled_sidecar_files(x)



import os
import re
import time
from concurrent.futures import ProcessPoolExecutor

import numpy as np
import pandas as pd
from aniposelib.cameras import CameraGroup
from fastapi import APIRouter, Depends
from pydantic import BaseModel

from litpose_app import deps
from litpose_app.config import Config
from litpose_app.routes.labeler import find_calibration_file, session_level_config_path
from litpose_app.routes.project import ProjectInfo
from litpose_app.tasks.extract_frames import MVLabelFile
from litpose_app.utils.fix_empty_first_row import fix_empty_first_row

router = APIRouter()

import logging

logger = logging.getLogger(__name__)


class BundleAdjustRequest(BaseModel):
    mvlabelfile: MVLabelFile
    sessionKey: str  # name of the session with the view stripped out


class BundleAdjustResponse(BaseModel):
    camList: list[str]
    """List of camera view names in order of the reprojection errors below."""

    oldReprojectionError: list[float]
    """List: one per camera."""

    newReprojectionError: list[float]
    """List: one per camera."""


@router.post("/app/v0/rpc/bundleAdjust")
def bundle_adjust(
    request: BundleAdjustRequest,
    project_info: ProjectInfo = Depends(deps.project_info),
    config: Config = Depends(deps.config),
) -> BundleAdjustResponse:
    with ProcessPoolExecutor(max_workers=1) as executor:
        fut = executor.submit(
            _bundle_adjust_impl,
            request,
            project_info,
            config,
        )
        result = fut.result()

    return BundleAdjustResponse.model_validate(result)


def _bundle_adjust_impl(
        request: BundleAdjustRequest, project_info: ProjectInfo, config: Config
):
    camera_group_toml_path = find_calibration_file(
        request.sessionKey, project_info, config
    )
    if camera_group_toml_path is None:
        raise FileNotFoundError(
            f"Could not find calibration file for {request.sessionKey}"
        )
    cg = CameraGroup.load(camera_group_toml_path)
    views = list(map(lambda c: c.name, cg.cameras))
    assert set(project_info.views) == set(views)

    def autoLabelSessionKey(framePath: str) -> str | None:
        parts = framePath.split("/")
        if len(parts) < 3:
            return False
        sessionViewNameWithDots = parts[-2]  # e.g. 05272019_fly1_0_R1C24_Cam-A_rot-ccw-0.06_sec

        def processPart(sessionViewName):
            """Mirrors frame.model.ts get autolabelSessionKey()"""
            # Replace view with *, e.g. 05272019_fly1_0_R1C24_*_rot-ccw-0.06_sec
            sessionkey_from_frame = re.sub(
                rf"({'|'.join([re.escape(_v) for _v in views])})", "*", sessionViewName
            )

            # View not in this token, so return identity.
            if '*' not in sessionkey_from_frame:
                return sessionkey_from_frame

            # Attempt to parse assuming - is the delimiter.
            parts_hyphenated = sessionkey_from_frame.split('-')
            if '*' in parts_hyphenated:
                return '-'.join(filter(lambda x: x != '*', parts_hyphenated))

            # Attempt to parse assuming _ is the delimiter.
            parts_underscored = sessionkey_from_frame.split('_')
            if '*' in parts_underscored:
                return '_'.join(filter(lambda x: x != '*', parts_underscored))

            # View present, but invalid delimiter: return None
            return None

        # Split on . and process each part.
        processedParts = list(map(processPart, sessionViewNameWithDots.split('.')))
        # If some part had * but without correct delimiters around it, return null.
        if None in processedParts:
            return None
        # Filter empty tokens after processPart (* got removed) and join by .
        return '.'.join(filter(lambda p: bool(p), processedParts))

    def is_of_current_session(imgpath: str):
        return autoLabelSessionKey(imgpath) == request.sessionKey

    # Group multiview csv files
    files_by_view = {v.viewName: v.csvPath for v in request.mvlabelfile.views}

    numpy_arrs: dict[str, np.ndarray] = {}  # view -> np.ndarray

    # Read DFs
    dfs_by_view = {}
    for view in views:
        csv = files_by_view[view]
        df = pd.read_csv(csv, header=[0, 1, 2], index_col=0)
        df = fix_empty_first_row(df)
        dfs_by_view[view] = df

    # Check that DFs are aligned
    index_values = dfs_by_view[views[0]].index.values
    firstview_framekeys = list(map(lambda s: s.replace(views[0], ""), index_values))
    for view in views:
        thisview_framekeys = list(
            map(lambda s: s.replace(view, ""), dfs_by_view[view].index.values)
        )
        if not firstview_framekeys == thisview_framekeys:
            print(f"Skipping {files_by_view[view]} because of misaligned indices")
            del files_by_view[view]
            continue

    # Filter to frames of current session
    for view in views:
        df = dfs_by_view[view]
        dfs_by_view[view] = df.loc[df.index.to_series().apply(is_of_current_session)]
        if len(dfs_by_view[view]) == 0:
            raise RuntimeError(
                f"Insufficient frames found after filtering for session {request.sessionKey}. Possible error in session extraction logic.")
    # Remove rows with NaN coordinates
    # Get indices of rows with NaN coordinates in any view
    nan_indices = set()
    for view in views:
        df = dfs_by_view[view]
        picked_columns = [c for c in df.columns if c[2] in ("x", "y")]
        nan_rows = df.loc[:, picked_columns].dropna().index.symmetric_difference(df.index)
        nan_indices.update(df.index.get_indexer(nan_rows))

    # Drop those indices from all views  
    for view in views:
        dfs_by_view[view] = dfs_by_view[view].drop(dfs_by_view[view].index[list(nan_indices)])
        if len(dfs_by_view[view]) == 0:
            raise RuntimeError(
                f"Insufficient frames found after dropping NaN rows for session {request.sessionKey}.")
    # Normalize columns: x, y alternating.
    for view in views:
        df = dfs_by_view[view]
        picked_columns = [c for c in df.columns if c[2] in ("x", "y")]
        assert len(picked_columns) % 2 == 0
        assert (
                picked_columns[::2][0][2] == "x"
                and len(set(map(lambda t: t[2], picked_columns[::2]))) == 1
        )
        assert (
                picked_columns[1::2][0][2] == "y"
                and len(set(map(lambda t: t[2], picked_columns[1::2]))) == 1
        )
        dfs_by_view[view] = df.loc[:, picked_columns]

    # Convert to numpy
    for view in views:
        df = dfs_by_view[view]
        nparr = df.to_numpy()
        # Convert from x, y alternating columns to just x, y columns
        # (bodyparts move from columns to rows).
        nparr = nparr.reshape(-1, 2)
        numpy_arrs[view] = nparr

    # Creates a CxNx2 np array for anipose
    p2ds = np.stack([numpy_arrs[v] for v in views])
    p3ds = cg.triangulate(p2ds)
    old_reprojection_error = cg.reprojection_error(p3ds, p2ds)
    cg.bundle_adjust(p2ds, verbose=True)
    p3ds = cg.triangulate(p2ds)
    new_reprojection_error = cg.reprojection_error(p3ds, p2ds)
    target_path = session_level_config_path(request.sessionKey, project_info, config)

    if target_path.exists():
        backup_path = (
                project_info.data_dir
                / config.CALIBRATION_BACKUPS_DIRNAME
                / target_path.with_suffix(f".{time.time_ns()}.toml").name
        )
        backup_path.parent.mkdir(parents=True, exist_ok=True)
        os.rename(target_path, backup_path)

    # Writes to session-level calibration file even if the
    # project level calibration file was used for bundle adjustment.
    session_level_calibration_path = session_level_config_path(
        request.sessionKey, project_info, config
    )
    session_level_calibration_path.parent.mkdir(parents=True, exist_ok=True)
    cg.dump(session_level_calibration_path)

    return {
        "camList": views,  # Add the camList
        "oldReprojectionError": np.linalg.norm(
            old_reprojection_error, axis=2
        )  # Change key to camelCase
        .sum(axis=1)
        .tolist(),
        "newReprojectionError": np.linalg.norm(
            new_reprojection_error, axis=2
        )  # Change key to camelCase
        .sum(axis=1)
        .tolist(),
    }
